"""Api testing of pyInels library."""
from pyinels.api import Api

from pyinels.exception import ApiException
from tests.const_test import (
    TEST_API_CLASS_NAMESPACE,
    TEST_DATA_SWITCH,
    TEST_HOST,
    TEST_PORT,
    TEST_RAW_DEVICES,
    TEST_ROOMS,
    TEST_VERSION
)

from unittest.mock import patch, Mock
from unittest import TestCase


class ApiTest(TestCase):
    """Class to test pyInels api library."""

    def setUp(self):
        """Setup all necessary instances nad mocks."""
        self.patches = [
            patch(f'{TEST_API_CLASS_NAMESPACE}.ping', return_value=True),
            patch(f'{TEST_API_CLASS_NAMESPACE}.read',
                  return_value={TEST_DATA_SWITCH['id']: 1}),
        ]

        self.proxy = Api(TEST_HOST, TEST_PORT, TEST_VERSION)

        # patching ping method in IneslBus3. It will be executed every test
        for p in self.patches:
            p.start()

    def tearDown(self):
        """Destroy all instances and mocks."""
        self.proxy = None
        patch.stopall()
        self.patches = None

    @patch(TEST_API_CLASS_NAMESPACE)
    def test_class_calling(self, mock_class):
        """Class instance test."""
        mock_class(TEST_HOST, TEST_PORT, TEST_VERSION)

        mock_class.assert_called()
        mock_class.assert_called_once()
        mock_class.assert_called_with(TEST_HOST, TEST_PORT, TEST_VERSION)

        self.assertEqual(mock_class.call_count, 1)

    @patch('xmlrpc.client.ServerProxy')
    def test_connection_failed(self, mock_server):
        """Test proxy connection."""
        mock_server.return_value = Mock()
        mock_server.side_effect = ApiException(
            'common_exception', 'Exception occur')

        ret = self.proxy.ping()
        self.assertEqual(True, ret)
        self.assertEqual(mock_server.call_count, 0)

    @patch(f'{TEST_API_CLASS_NAMESPACE}.ping')
    def test_ping_failed(self, mock_method):
        """Test ping method."""
        mock_method.return_value = False
        ret = self.proxy.ping()

        self.assertEqual(mock_method.call_count, 1)
        self.assertFalse(ret)

    @patch(f'{TEST_API_CLASS_NAMESPACE}.getPlcIp')
    def test_getPlcIp_success(self, mock_method):
        """Test Ip address of the PLC."""
        mock_method.return_value = "192.168.2.10"

        ret = self.proxy.getPlcIp()

        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(ret, "192.168.2.10")

    @patch(f'{TEST_API_CLASS_NAMESPACE}.getRooms')
    def test_getRoomsRaw_list(self, mock_method):
        """Test list of rooms defined on Connection server."""
        mock_method.return_value = TEST_ROOMS

        ret = self.proxy.getRooms()

        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(len(ret), 6)
        self.assertEqual(ret[1], 'First floor')

    @patch(f'{TEST_API_CLASS_NAMESPACE}.getRoomDevicesRaw')
    def test_getRoomDevices_list(self, mock_method):
        """Test list of all devices in room."""
        mock_method.return_value = TEST_RAW_DEVICES
        raw = self.proxy.getRoomDevicesRaw('room')

        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(len(raw), len(mock_method.return_value))

        # this is the way how to mock private method
        with patch.object(
            self.proxy,
            '_Api__readDeviceData',
            return_value={'Doors_Garage': 0}
        ):
            with patch.object(self.proxy, 'ping', return_value=True):
                obj_list = self.proxy.getRoomDevices('room')

                device = obj_list[0]

                self.assertGreater(len(obj_list), 0)
                self.assertEqual(device.id, 'Doors_Garage')
                self.assertEqual(device.title, 'Doors')

                patch.stopall()
                device_value = self.proxy.read([device.id])
                self.assertEqual(device_value, {'Doors_Garage': 0})



"""Const of iNels BUS."""

# Inels attributes
ATTR_DOWN = "Down"
ATTR_GROUP = "Group"
ATTR_ID = "Id"
ATTR_READ_ONLY = "Read_only"
ATTR_RELE = "Rele"
ATTR_SWITCH_OFF = "0"
ATTR_SWITCH_ON = "1"
ATTR_TEMP = "Temp"
ATTR_TEMP_SET = "Temp_set"
ATTR_TITLE = "Title"
ATTR_TYPE = "Type"
ATTR_UP = "Up"

# Internal device attributes
ATTR_SWITCH = "switch"
ATTR_LIGHT = "light"

# Device types dictionary with all determined typs translated to better
# understand
DEVICE_TYPE_DICT = {
    "undefined": "undefined",
    "lights": "light",
    "on_off": "switch",
    "shutters": "shutter",
    "heat-control": "therm",
    "heating": "heating",
    "garage": "garage"
}

INELS_BUS_ATTR_DICT = {
    "Id": "inels",
    "Title": "name",
    "Type": "type",
    "Temp": "therm",
    "Rele": "rele",
    "Temp_set": "stateth",
    "Down": "down",
    "Up": "up",
    "Read_only": "read_only",
    "Group": "group"
}

NAME = "pyinels"

RANGE_BLIND = (0, 100)
RANGE_BRIGHTNESS = (0.0, 255.0)

SUPPORT_BRIGHTNESS = 1
SUPPORT_COLOR_TEMP = 4
SUPPORT_HEX_COLOR = 8
SUPPORT_RGB_COLOR = 2

VERSION = "0.4.3"



"""Inels switch class for iNels BUS."""

from pyinels.const import (
    ATTR_SWITCH_ON
)


class pySwitch:
    """Switch class based on InelsDevice."""

    def __init__(self, device):
        """Initialize of object InelsSwitch."""
        self.__device = device

    @property
    def state(self):
        """Return the state of the switch."""
        if (self.__device.value is None):
            self.__device.observe()

        val = str(self.__device.value)
        attr = ATTR_SWITCH_ON

        # it is expression of ternary operator
        return (True if val == attr else False)

    @property
    def name(self):
        """Name of the device."""
        return self.__device.title

    @property
    def unique_id(self):
        """Unique id of the device."""
        return self.__device.id

    def turn_off(self):
        """Turn the switch off."""
        self.__device.set_value(0)

    def turn_on(self):
        """Turn the switch on."""
        self.__device.set_value(1)

    def __repr__(self):
        """Object representation."""
        state = "on" if self.state else "off"
        return "<Switch #{} - " \
            "title: {}, " \
            "state: {}" \
            ">".format(self.__device.id, self.__device.title, state)



"""Unit testing of iNels switch library."""
from pyinels.const import ATTR_SWITCH

from tests.const_test import (
    TEST_API_CLASS_NAMESPACE,
    TEST_HOST,
    TEST_PORT,
    TEST_RAW_DEVICES,
    TEST_VERSION
)

from pyinels.api import Api
from pyinels.device.pySwitch import pySwitch

from unittest.mock import patch
from unittest import TestCase


class PySwitchTest(TestCase):
    """Class to test iNels switch library."""

    def setUp(self):
        """Setup all necesary instances and mocks."""
        self.patches = [
            patch(f'{TEST_API_CLASS_NAMESPACE}.ping', return_value=True),
            patch(f'{TEST_API_CLASS_NAMESPACE}.getRoomDevicesRaw',
                  return_value=TEST_RAW_DEVICES)
        ]

        for p in self.patches:
            p.start()

        self.api = Api(TEST_HOST, TEST_PORT, TEST_VERSION)
        switches = [device for device in self.api.getRoomDevices(
            'garage') if device.type == ATTR_SWITCH]

        self.device = switches[0]

    def tearDown(self):
        """Destroy all instances and mocks."""
        self.api = None
        self.device = None
        patch.stopall()
        self.patches = None

    def test_state(self):
        """Test the state of the pySwitch."""
        with patch.object(self.api, '_Api__readDeviceData',
                          return_value={'ZA_01_GARAGE': 0}):
            p = pySwitch(self.device)

            # the switch at the beggining should be turned off
            self.assertFalse(p.state)

            with patch.object(self.api, '_Api__writeValues',
                              return_value=None):
                # tur on the switch
                p.turn_on()

                self.assertTrue(p.state)



"""Class Api resources returned from the iNels BUS."""
import logging

from pyinels.exception import ApiClassTypeException

from pyinels.const import (
    ATTR_DOWN,
    ATTR_GROUP,
    ATTR_ID,
    ATTR_READ_ONLY,
    ATTR_RELE,
    ATTR_TEMP,
    ATTR_TEMP_SET,
    ATTR_TITLE,
    ATTR_TYPE,
    ATTR_UP,
    INELS_BUS_ATTR_DICT
)

_LOGGER = logging.getLogger(__name__)


class ApiResource:
    """Object returned from the iNels BUS."""

    @property
    def id(self):
        """Id of the object."""
        if INELS_BUS_ATTR_DICT.get(ATTR_ID) not in self.__json:
            return None
        return self.__json[INELS_BUS_ATTR_DICT.get(ATTR_ID)]

    @property
    def title(self):
        """Name of the object."""
        return self.__json[INELS_BUS_ATTR_DICT.get(ATTR_TITLE)]

    @property
    def type(self):
        """Type of the object."""
        return self.__json[INELS_BUS_ATTR_DICT.get(ATTR_TYPE)]

    @property
    def temperature(self):
        """Temperature of the object."""
        if INELS_BUS_ATTR_DICT.get(ATTR_TEMP) not in self.__json:
            return None
        return self.__json[INELS_BUS_ATTR_DICT.get(ATTR_TEMP)]

    @property
    def temperature_set(self):
        """Temperature of the object."""
        if INELS_BUS_ATTR_DICT.get(ATTR_TEMP_SET) not in self.__json:
            return None
        return self.__json[INELS_BUS_ATTR_DICT.get(ATTR_TEMP_SET)]

    @property
    def rele(self):
        """Rele of the object."""
        if INELS_BUS_ATTR_DICT.get(ATTR_RELE) not in self.__json:
            return None
        return self.__json[INELS_BUS_ATTR_DICT.get(ATTR_RELE)]

    @property
    def read_only(self):
        """Read only of the object."""
        if INELS_BUS_ATTR_DICT.get(ATTR_READ_ONLY) not in self.__json:
            return None
        return self.__json[INELS_BUS_ATTR_DICT.get(ATTR_READ_ONLY)]

    @property
    def down(self):
        """Shutter down of the object."""
        if INELS_BUS_ATTR_DICT.get(ATTR_DOWN) not in self.__json:
            return None
        return self.__json[INELS_BUS_ATTR_DICT.get(ATTR_DOWN)]

    @property
    def up(self):
        """Shutter up of the object."""
        if INELS_BUS_ATTR_DICT.get(ATTR_UP) not in self.__json:
            return None
        return self.__json[INELS_BUS_ATTR_DICT.get(ATTR_UP)]

    @property
    def group(self):
        """Group from where the object is loaded."""
        if INELS_BUS_ATTR_DICT.get(ATTR_GROUP) not in self.__json:
            return None
        return self.__json[INELS_BUS_ATTR_DICT.get(ATTR_GROUP)]

    @property
    def value(self):
        """Value of the returned object."""
        if hasattr(self, '_ApiResource__value'):
            return self.__value

        return None

    def __init__(self, json, api):
        """Initializer of the Api resource."""
        self.__json = json
        self.__api = api

    def observe(self, options=None):
        """Read the current value of the device."""
        try:
            if options:
                if not isinstance(options, Observe):
                    raise ApiClassTypeException(
                        '500', f"""Device.observe options
                        param has bad type. {type(options)}.
                        Should by Observe.""")

            raw = self.__api.read([self.id])
            self.__value = raw[self.id]

            if options:
                options.callback(self)

            return self.__value
        except ApiClassTypeException as ex:
            raise ex
        except Exception:
            if options:
                options.err_callback()
            # this is the situation when the proxy is
            # probably not available, then we are going to
            # se the value to None
            self.__value = None

        return None

    def set_value(self, value):
        """Set value to the device."""

        # initialize __value attribute
        if hasattr(self, '_ApiResource__value') is False:
            self.__value = 0

        curr_int = isinstance(self.__value, int)
        new_int = isinstance(value, int)

        if (self.__value != value or curr_int is not new_int):
            self.__value = value
            self.__api.write(self, value)

    @property
    def is_available(self):
        """Device availability property."""
        # first test when device has any value
        if hasattr(self, '_ApiResource__value'):
            return True
        else:
            # if not then try to observer
            self.observe()
        # when nothing change then device is not available
        return False if self.__value is None else True


class Observe:
    """Class defined observe option object."""

    @property
    def callback(self):
        """Callback property."""
        return self.__callback

    @property
    def err(self):
        """Error property."""
        return self.__err_callback

    def __init__(self, callback, err):
        """Initialize Observe object."""
        self.__callback = callback
        self.__err_callback = err



"""Inels light class for iNels BUS."""

from pyinels.const import (
    RANGE_BRIGHTNESS
)

MIN_RANGE = RANGE_BRIGHTNESS[0]
MAX_RANGE = RANGE_BRIGHTNESS[1]


class pyLight:
    """Inels class based on InelsDevice."""

    def __init__(self, device):
        """Initialize object."""
        self.__device = device

    @property
    def state(self):
        """State of the light."""
        if self.__device.value is None:
            self.__device.observe()

        return True if self.__device.value > 0 else False

    @property
    def name(self):
        """Name of the light."""
        return self.__device.title

    @property
    def unique_id(self):
        """Unique id of the device."""
        return self.__device.id

    @property
    def has_brightness(self):
        """Supports brightness."""
        if self.__device.value is None:
            self.__device.observe()

        return isinstance(self.__device.value, float)

    def set_brightness(self, value):
        """Set brightness of the light."""
        if (self.has_brightness
                and (value > MIN_RANGE and value <= MAX_RANGE)):
            self.__device.set_value(value)

    def turn_off(self):
        """Turn off the light."""
        self.__device.set_value(0)

    def turn_on(self):
        """Turn on the light."""
        if self.has_brightness is True:
            # when the light has brightness then set max range for turn on
            # and min range for turn off
            self.__device.set_value(
                MIN_RANGE if self.state is True else MAX_RANGE)
        else:
            # set device value to 0 when turn off and to 1 when turn on
            self.__device.set_value(1)



"""Resource testing."""
from unittest import TestCase
from unittest.mock import patch

from pyinels.api import Api
from pyinels.api.resources import ApiResource
from tests.const_test import (
    TEST_API_CLASS_NAMESPACE,
    TEST_API_NAMESPACE,
    TEST_HOST,
    TEST_PORT,
    TEST_RAW_DEVICES,
    TEST_RESOURCE_SWITCH,
    TEST_VERSION
)


class ResourceTest(TestCase):
    """Class to test resource of api iNels BUS."""

    def setUp(self):
        self.api = Api(TEST_HOST, TEST_PORT, TEST_VERSION)

        self.patches = [
            patch(f'{TEST_API_CLASS_NAMESPACE}.ping', return_value=True),
            patch(f'{TEST_API_CLASS_NAMESPACE}.getRoomDevicesRaw',
                  return_value=TEST_RAW_DEVICES)
        ]

        for p in self.patches:
            p.start()

        self.res_list = self.api.getRoomDevices('garage')

        self.garage_door = self.res_list[0]

    def tearDown(self):
        """Destroy all instances and mocks."""
        self.api = None
        self.res_list = None
        self.garage_door = None
        patch.stopall()
        self.patches = None

    @patch(f'{TEST_API_NAMESPACE}.resources.ApiResource')
    def test_load_resource_object(self, mock_class):
        """Test to load resource object."""
        mock_class(TEST_RESOURCE_SWITCH, self.api)

        mock_class.assert_called()
        mock_class.assert_called_once()

        self.assertEqual(mock_class.call_count, 1)

        res = ApiResource(TEST_RESOURCE_SWITCH, self.api)

        self.assertIsInstance(res, ApiResource)

        self.assertEqual(TEST_RESOURCE_SWITCH['type'], res.type)
        # inels is raw data id from iNels BUS
        self.assertEqual(TEST_RESOURCE_SWITCH['inels'], res.id)
        self.assertEqual(TEST_RESOURCE_SWITCH['name'], res.title)
        self.assertEqual(TEST_RESOURCE_SWITCH['read_only'], res.read_only)
        # should be none, because it does not get the observe
        self.assertIsNone(res.value)

    @patch(f'{TEST_API_CLASS_NAMESPACE}.read')
    def test_observe(self, mock_room_devices):
        """Test the observe method of the Api resources. It should touche
        the iNels BUS."""
        mock_room_devices.return_value = {'Doors_Garage': 0}

        self.assertEqual(len(self.res_list), 4)
        self.assertEqual(self.garage_door.title, 'Doors')

        value = self.garage_door.observe()

        self.assertEqual(value, 0)

        self.assertIsNotNone(self.garage_door.value)
        self.assertEqual(self.garage_door.value, 0)

    def test_set_value(self):
        """Test set value to the iNels BUS."""
        with patch.object(self.api, '_Api__writeValues', return_value=None):
            # set int
            self.garage_door.set_value(1)
            self.assertEqual(self.garage_door.value, 1)

            # change int to another value
            self.garage_door.set_value(0)
            self.assertEqual(self.garage_door.value, 0)

            # change to float with different value
            self.garage_door.set_value(25.0)
            self.assertEqual(self.garage_door.value, 25.0)

            # change to int with same value
            self.garage_door.set_value(25)
            self.assertEqual(self.garage_door.value, 25)

    def test_is_available(self):
        """Test when the resource object is available."""
        with patch.object(self.api, '_Api__writeValues', return_value=None):
            # not set value of the ApiResource object then is not available
            self.assertFalse(self.garage_door.is_available)

            # set the value of the ApiResource then it should be available
            self.garage_door.set_value(1)
            self.assertTrue(self.garage_door.is_available)



"""Const of iNels BUS tests."""
TEST_API_NAMESPACE = "pyinels.api"
TEST_API_CLASS_NAMESPACE = "pyinels.api.Api"

TEST_DATA_THERM = {
    'title': 'Valve in bedroom',
    'temp': 'Bedroom_temerature',
    'rele': 'DIGITALGROUP_0001',
    'read_only': 'no',
    'temp_set': 'Bedroom_stateth'
}

TEST_DATA_LIGHT = {
    'title': 'Kitchen main light',
    'id': 'KITCHEN_MAIN_LIGHT',
    'read_only': 'no'
}

TEST_DATA_SWITCH = {
    'title': 'Kitchen kettel switch',
    'id': 'KITCHEN_KETTLE_SWITCH',
    'read_only': 'no'
}

TEST_DATA_SHUTTER = {
    'title': 'Kitchen shuuter',
    'up': 'KITCHEN_SHUUTER_1_UP',
    'down': 'KITCHEN_SHUUTER_1_DOWN',
    'read_only': 'no'
}

TEST_DATA_GARAGE = {
    'title': 'Garage door',
    'id': 'GARAGE_DOOR',
    'read_only': 'no'
}

TEST_DATA_HEATING = {
    'title': 'Holiday',
    'id': 'HOLIDAY_HEATING',
    'read_only': 'no'
}

TEST_HOST = "http://localhost"

TEST_INELS_DEVICE_NAMESPACE = "pyinels.device.Device"

TEST_PORT = "1000"

TEST_RESOURCE_SWITCH = {
    'column': '0',
    'group': 'basement',
    'inels': 'ZA_14_basement',
    'name': 'Office switch',
    'read_only': 'no',
    'row': '1',
    'type': 'switch'
}

TEST_RAW_DEVICES = """garage:
name="Doors" column="0" inels="Doors_Garage" read_only="no" row="1"
lights:
name="Main light" column="0" inels="SV_12_Garage" read_only="no" row="0"
name="Wall light" column="1" inels="SV_Wall_Garage" read_only="no" row="0"
on_off:
name="Main switch" column="2" inels="ZA_01_GARAGE" read_only="no" row="0"
"""
TEST_ROOMS = ['Basement', 'First floor', 'Front yard',
              'Back yard', 'Attic', 'Garrage']

TEST_VERSION = "CU3"



"""Common device class for iNels BUS."""
from pyinels.api.resources import ApiResource


class Device(ApiResource):
    def __repr__(self):
        """Object representation."""
        return "<{} - {} type({})>".format(self.id, self.title, self.type)



# """Unit testing of iNels light library."""

# from tests.const_test import (
#     TEST_DATA_LIGHT,
#     TEST_HOST,
#     TEST_INELS_BUS3_NAMESPACE,
#     TEST_INELS_DEVICE_NAMESPACE,
#     TEST_PORT
# )

# from pyinels.api import Api
# from pyinels.device import (
#     InelsDevice,
#     DeviceType
# )

# from pyinels.device.pyLight import pyLight

# from unittest.mock import patch
# from unittest import TestCase


# class pyLightTest(TestCase):
#     """Class to test iNels light library."""

#     def setUp(self):
#         """Setup all necesary instances and mocks."""
#         self.patches = [
#             patch(f'{TEST_INELS_BUS3_NAMESPACE}.ping', return_value=True),
#             patch(f'{TEST_INELS_BUS3_NAMESPACE}.read',
#                   return_value={TEST_DATA_LIGHT['id']: 1}),
#             patch(f'{TEST_INELS_DEVICE_NAMESPACE}._write', return_value=None)
#         ]

#         proxy = InelsBus3(TEST_HOST, TEST_PORT)
#         device = InelsDevice(TEST_DATA_LIGHT['name'],
#                              TEST_DATA_LIGHT['id'],
#                              DeviceType.LIGHT, proxy)

#         self.light = pyLight(device)
#         # patching ping method in IneslBus3. It will be executed every test
#         for p in self.patches:
#             p.start()

#     def tearDown(self):
#         """Destroy all instances and mocks."""
#         self.light = None
#         patch.stopall()

#     def test_initialize_light(self):
#         """Initialize light test."""
#         self.assertEqual(TEST_DATA_LIGHT['id'], self.light.device.id)

#     def test_state_light(self):
#         """Test state of the light."""
#         state = self.light.state
#         self.assertEqual(state, True)

#     def test_has_not_brightness(self):
#         """Test that the light has no brightness feature."""
#         brightness_support = self.light.has_brightness
#         self.assertFalse(brightness_support)

#     def test_has_brightness(self):
#         """Test that the light has brightness feature."""
#         self.light.device.set_value(25.0)
#         brightness_support = self.light.has_brightness

#         self.assertTrue(brightness_support)

#     def test_set_brightness(self):
#         """Test set brightness of the light."""
#         value = self.light.device.observe()
#         # at the begining the light has no brightness support

#         self.light.set_brightness(0.0)
#         self.assertFalse(self.light.has_brightness)
#         # no change
#         self.assertEqual(value, self.light.device.value)

#         # set brightness support to the light
#         self.light.device.set_value(0.0)
#         self.assertTrue(self.light.has_brightness)

#         self.light.set_brightness(25.0)
#         self.assertNotEqual(value, self.light.device.value)
#         self.assertEqual(self.light.device.value, 25.0)

#         self.light.set_brightness(256.0)
#         # brightness is out of range, then no changes
#         self.assertNotEqual(self.light.device.value, 256.0)
#         self.assertEqual(self.light.device.value, 25.0)

#         # remove brightness support
#         self.light.device.set_value(0)
#         self.assertFalse(self.light.has_brightness)
#         self.light.set_brightness(25.0)
#         self.assertNotEqual(self.light.device.value, 25.0)
#         self.assertEqual(self.light.device.value, 0)

#     def test_set_state(self):
#         """Test set state of the light."""
#         # set the light to turn off without brightness support
#         # for testing purposes
#         self.light.device.set_value(0)

#         self.light.set_state()
#         self.assertTrue(self.light.state)
#         self.assertEqual(self.light.device.value, 1)

#         # turn off the ligth
#         self.light.set_state()
#         self.assertFalse(self.light.state)
#         self.assertEqual(self.light.device.value, 0)

#         # set the brightness support
#         self.light.device.set_value(0.0)
#         # turn on the ligt wit brightness
#         self.light.set_state()
#         self.assertTrue(self.light.state)
#         self.assertEqual(self.light.device.value, 255.0)
